"""
This module tests the various reporters included in the parmed package
"""
import numpy as np
import os
from io import StringIO
from parmed import unit as u, load_file
from parmed.amber import AmberParm, AmberMdcrd, AmberAsciiRestart, NetCDFTraj, NetCDFRestart
from parmed.openmm.reporters import (
    NetCDFReporter, MdcrdReporter, ProgressReporter, RestartReporter, StateDataReporter,
    EnergyMinimizerReporter, _format_time
)
import sys
import unittest
from utils import mm, app, has_openmm, CPU, FileIOTestCase, HAS_GROMACS

@unittest.skipUnless(has_openmm, "Cannot test without OpenMM")
class TestStateDataReporter(FileIOTestCase):

    def setUp(self):
        super().setUp()
        self.amber_gas = AmberParm(self.get_fn('ash.parm7'), self.get_fn('ash.rst7'))

    def test_state_data_reporter(self):
        """ Test StateDataReporter with various options """
        system = self.amber_gas.createSystem()
        integrator = mm.LangevinIntegrator(300*u.kelvin, 5.0/u.picoseconds, 1.0*u.femtoseconds)
        sim = app.Simulation(self.amber_gas.topology, system, integrator, platform=CPU)
        sim.context.setPositions(self.amber_gas.positions)
        f = open(self.get_fn('akma5.dat', written=True), 'w')
        sim.reporters.extend([
            StateDataReporter(self.get_fn('akma1.dat', written=True), 10),
            StateDataReporter(self.get_fn('akma2.dat', written=True), 10,
                              time=False, potentialEnergy=False,
                              kineticEnergy=False, totalEnergy=False,
                              temperature=False),
            StateDataReporter(self.get_fn('akma3.dat', written=True), 10, volume=True, density=True),
            StateDataReporter(self.get_fn('akma4.dat', written=True), 10, separator='\t'),
            StateDataReporter(self.get_fn('units.dat', written=True), 10, volume=True, density=True,
                              energyUnit=u.kilojoules_per_mole, volumeUnit=u.nanometers**3),
            StateDataReporter(f, 10)
        ])
        sim.step(500)
        f.close()

        # Now open all of the reporters and check that the information in there
        # is what we expect it to be
        akma1 = open(self.get_fn('akma1.dat', written=True), 'r')
        akma2 = open(self.get_fn('akma2.dat', written=True), 'r')
        akma3 = open(self.get_fn('akma3.dat', written=True), 'r')
        akma4 = open(self.get_fn('akma4.dat', written=True), 'r')
        akma5 = open(self.get_fn('akma5.dat', written=True), 'r')
        units = open(self.get_fn('units.dat', written=True), 'r')
        # AKMA 1 first
        header = akma1.readline().strip()[1:].split(',')
        self.assertEqual(len(header), 6)
        for i, label in enumerate(
            ('Step', 'Time', 'Potential Energy', 'Kinetic Energy', 'Total Energy', 'Temperature')
        ):
            self.assertTrue(label in header[i])
        for i, line in enumerate(akma1):
            words = line.replace('\n', '').split(',')
            self.assertEqual(i*10+10, int(words[0])) # step
        akma1.close()
        # AKMA 2
        header = akma2.readline().strip()[1:].split(',')
        self.assertEqual(len(header), 1)
        self.assertTrue('Step' in header[0])
        for i, line in enumerate(akma2):
            self.assertEqual(int(line.replace('\n', '').split(',')[0]), 10*i+10)
        akma2.close()
        # AKMA 3 -- save energies so we can compare to the file with different
        # units
        header = akma3.readline().strip()[1:].split(',')
        self.assertEqual(len(header), 8)
        for i, label in enumerate(('Step', 'Time', 'Potential Energy',
                                   'Kinetic Energy', 'Total Energy',
                                   'Temperature', 'Box Volume', 'Density')):
            self.assertTrue(label in header[i])
        akma_energies = [[0.0 for i in range(8)] for j in range(50)]
        for i, line in enumerate(akma3):
            words = line.replace('\n', '').split(',')
            akma_energies[i][0] = int(words[0])
            for j in range(1, 8):
                akma_energies[i][j] = float(words[j])
        akma3.close()
        # AKMA 4 -- tab delimiter
        header = akma4.readline().strip()[1:].split('\t')
        self.assertEqual(len(header), 6)
        for i, label in enumerate(('Step', 'Time', 'Potential Energy',
                                   'Kinetic Energy', 'Total Energy',
                                   'Temperature')):
            self.assertTrue(label in header[i])
        for i, line in enumerate(akma4):
            words = line.replace('\n', '').split('\t')
            self.assertEqual(i*10+10, int(words[0])) # step
        akma4.close()
        # AKMA 5 -- write to open file handle
        header = akma5.readline().strip()[1:].split(',')
        self.assertEqual(len(header), 6)
        for i, label in enumerate(('Step', 'Time', 'Potential Energy',
                                   'Kinetic Energy', 'Total Energy',
                                   'Temperature')):
            self.assertTrue(label in header[i])
        for i, line in enumerate(akma5):
            words = line.replace('\n', '').split(',')
            self.assertEqual(i*10+10, int(words[0])) # step
        akma5.close()
        # UNITS -- compare other units
        ene = u.kilojoule_per_mole.conversion_factor_to(u.kilocalorie_per_mole)
        volume = u.nanometers.conversion_factor_to(u.angstroms)**3
        conversions = [1, 1, ene, ene, ene, 1, volume, 1]
        headers = units.readline().strip()[1:].split(',')
        self.assertEqual(len(headers), 8)
        for i, line in enumerate(units):
            words = line.replace('\n', '').split(',')
            self.assertEqual(int(words[0]), akma_energies[i][0])
            for j in range(1, 8):
                self.assertAlmostEqual(float(words[j])*conversions[j],
                                       akma_energies[i][j],
                                       places=5)
        units.close()

    def test_progress_reporter(self):
        """ Test ProgressReporter with various options """
        self.assertRaises(ValueError, lambda: ProgressReporter(sys.stdout, 1, 5))
        system = self.amber_gas.createSystem()
        integrator = mm.LangevinIntegrator(300*u.kelvin, 5.0/u.picoseconds,
                                           1.0*u.femtoseconds)
        sim = app.Simulation(self.amber_gas.topology, system, integrator, platform=CPU)
        sim.context.setPositions(self.amber_gas.positions)
        sim.reporters.append(
            ProgressReporter(self.get_fn('progress_reporter.dat', written=True), 10,
                             500, step=True, time=True, potentialEnergy=True,
                             kineticEnergy=True, totalEnergy=True,
                             temperature=True, volume=True, density=True,
                             systemMass=None)
        )
        sim.step(500)
        self.assertEqual(len(os.listdir(self._temporary_directory.name)), 1)
        text = open(self.get_fn('progress_reporter.dat', written=True), 'r').read()
        self.assertTrue('Estimated time to completion' in text)
        self.assertTrue('Total Energy' in text)
        self.assertTrue('Potential Energy' in text)
        self.assertTrue('Kinetic Energy' in text)
        self.assertTrue('Temperature' in text)

@unittest.skipUnless(has_openmm, "Cannot test without OpenMM")
class TestTrajRestartReporter(FileIOTestCase):

    def setUp(self):
        super().setUp()
        self.amber_gas = AmberParm(self.get_fn('ash.parm7'), self.get_fn('ash.rst7'))

    def test_reporters(self):
        """ Test NetCDF and ASCII restart and trajectory reporters (no PBC) """
        with self.assertRaises(ValueError):
            NetCDFReporter(self.get_fn('blah', written=True), 1, crds=False)
        with self.assertRaises(ValueError):
            MdcrdReporter(self.get_fn('blah', written=True), 1, crds=False)
        with self.assertRaises(ValueError):
            MdcrdReporter(self.get_fn('blah', written=True), 1, crds=True, vels=True)
        system = self.amber_gas.createSystem()
        integrator = mm.LangevinIntegrator(300*u.kelvin, 5.0/u.picoseconds, 1.0*u.femtoseconds)
        sim = app.Simulation(self.amber_gas.topology, system, integrator, platform=CPU)
        sim.context.setPositions(self.amber_gas.positions)
        sim.reporters.extend([
            NetCDFReporter(self.get_fn('traj1.nc', written=True), 10),
            NetCDFReporter(self.get_fn('traj2.nc', written=True), 10, vels=True),
            NetCDFReporter(self.get_fn('traj3.nc', written=True), 10, frcs=True),
            NetCDFReporter(self.get_fn('traj4.nc', written=True), 10, vels=True, frcs=True),
            NetCDFReporter(self.get_fn('traj5.nc', written=True), 10, crds=False, vels=True),
            NetCDFReporter(self.get_fn('traj6.nc', written=True), 10, crds=False, frcs=True),
            NetCDFReporter(self.get_fn('traj7.nc', written=True), 10, crds=False, vels=True, frcs=True),
            MdcrdReporter(self.get_fn('traj1.mdcrd', written=True), 10),
            MdcrdReporter(self.get_fn('traj2.mdcrd', written=True), 10, crds=False, vels=True),
            MdcrdReporter(self.get_fn('traj3.mdcrd', written=True), 10, crds=False, frcs=True),
            RestartReporter(self.get_fn('restart.ncrst', written=True), 10, write_multiple=True, netcdf=True),
            RestartReporter(self.get_fn('restart.rst7', written=True), 10),
        ])
        sim.step(500)
        for reporter in sim.reporters:
            reporter.finalize()

        self.assertEqual(len(os.listdir(self._temporary_directory.name)), 61)
        ntraj = [NetCDFTraj.open_old(self.get_fn('traj1.nc', written=True)),
                 NetCDFTraj.open_old(self.get_fn('traj2.nc', written=True)),
                 NetCDFTraj.open_old(self.get_fn('traj3.nc', written=True)),
                 NetCDFTraj.open_old(self.get_fn('traj4.nc', written=True)),
                 NetCDFTraj.open_old(self.get_fn('traj5.nc', written=True)),
                 NetCDFTraj.open_old(self.get_fn('traj6.nc', written=True)),
                 NetCDFTraj.open_old(self.get_fn('traj7.nc', written=True))]
        atraj = [
            AmberMdcrd(self.get_fn('traj1.mdcrd', written=True), self.amber_gas.ptr('natom'), hasbox=False, mode='r'),
            AmberMdcrd(self.get_fn('traj2.mdcrd', written=True), self.amber_gas.ptr('natom'), hasbox=False, mode='r'),
            AmberMdcrd(self.get_fn('traj3.mdcrd', written=True), self.amber_gas.ptr('natom'), hasbox=False, mode='r'),
        ]
        for traj in ntraj:
            self.assertEqual(traj.frame, 50)
            self.assertEqual(traj.Conventions, 'AMBER')
            self.assertEqual(traj.ConventionVersion, '1.0')
            self.assertEqual(traj.application, 'AmberTools')
            self.assertEqual(traj.program, 'ParmEd')
            self.assertFalse(traj.hasbox)
        self.assertTrue(ntraj[0].hascrds)
        self.assertFalse(ntraj[0].hasvels)
        self.assertFalse(ntraj[0].hasfrcs)
        self.assertTrue(ntraj[1].hascrds)
        self.assertTrue(ntraj[1].hasvels)
        self.assertFalse(ntraj[1].hasfrcs)
        self.assertTrue(ntraj[2].hascrds)
        self.assertFalse(ntraj[2].hasvels)
        self.assertTrue(ntraj[2].hasfrcs)
        self.assertTrue(ntraj[3].hascrds)
        self.assertTrue(ntraj[3].hasvels)
        self.assertTrue(ntraj[3].hasfrcs)
        self.assertFalse(ntraj[4].hascrds)
        self.assertTrue(ntraj[4].hasvels)
        self.assertFalse(ntraj[4].hasfrcs)
        self.assertFalse(ntraj[5].hascrds)
        self.assertFalse(ntraj[5].hasvels)
        self.assertTrue(ntraj[5].hasfrcs)
        self.assertFalse(ntraj[6].hascrds)
        self.assertTrue(ntraj[6].hasvels)
        self.assertTrue(ntraj[6].hasfrcs)
        for i in (0, 2, 3, 4, 5, 6):
            ntraj[i].close() # still need the 2nd
        for traj in atraj:
            traj.close()
        # Now test the NetCDF restart files
        fn = self.get_fn('restart.ncrst.%d', written=True)
        for i, j in enumerate(range(10, 501, 10)):
            ncrst = NetCDFRestart.open_old(fn % j)
            self.assertEqual(ncrst.coordinates.shape, (1, 25, 3))
            self.assertEqual(ncrst.velocities.shape, (1, 25, 3))
            np.testing.assert_allclose(ncrst.coordinates[0], ntraj[1].coordinates[i])
            np.testing.assert_allclose(ncrst.velocities[0], ntraj[1].velocities[i], rtol=1e-6)
        # Now test the ASCII restart file
        f = AmberAsciiRestart(self.get_fn('restart.rst7', written=True), 'r')
        # Compare to ncrst and make sure it's the same data
        np.testing.assert_allclose(ncrst.coordinates, f.coordinates, atol=1e-3)
        np.testing.assert_allclose(ncrst.velocities, f.velocities, rtol=1e-3)
        # Make sure the EnergyMinimizerReporter does not fail
        f = StringIO()
        rep = EnergyMinimizerReporter(f)
        rep.report(sim, frame=10)
        rep.finalize()

    @unittest.skipUnless(HAS_GROMACS, 'Cannot test without GROMACS')
    def test_reporters_pbc(self):
        """ Test NetCDF and ASCII restart and trajectory reporters (w/ PBC) """
        systemsolv = load_file(self.get_fn('ildn.solv.top'), xyz=self.get_fn('ildn.solv.gro'))
        system = systemsolv.createSystem(nonbondedMethod=app.PME,
                                         nonbondedCutoff=8*u.angstroms)
        integrator = mm.LangevinIntegrator(300*u.kelvin, 5.0/u.picoseconds,
                                           1.0*u.femtoseconds)
        sim = app.Simulation(systemsolv.topology, system, integrator, CPU)
        sim.context.setPositions(systemsolv.positions)
        sim.reporters.extend([
                NetCDFReporter(self.get_fn('traj.nc', written=True), 1, vels=True, frcs=True),
                MdcrdReporter(self.get_fn('traj.mdcrd', written=True), 1),
                RestartReporter(self.get_fn('restart.ncrst', written=True), 1, netcdf=True),
                RestartReporter(self.get_fn('restart.rst7', written=True), 1),
                StateDataReporter(self.get_fn('state.o', written=True), 1, volume=True, density=True, systemMass=1)
        ])
        sim.step(5)
        for reporter in sim.reporters: reporter.finalize()
        ntraj = NetCDFTraj.open_old(self.get_fn('traj.nc', written=True))
        atraj = AmberMdcrd(self.get_fn('traj.mdcrd', written=True), len(systemsolv.atoms), True, mode='r')
        nrst = NetCDFRestart.open_old(self.get_fn('restart.ncrst', written=True))
        arst = AmberAsciiRestart(self.get_fn('restart.rst7', written=True), 'r')
        self.assertEqual(ntraj.frame, 5)
        self.assertEqual(atraj.frame, 5)
        self.assertTrue(ntraj.hasvels)
        self.assertTrue(ntraj.hasfrcs)
        for i in range(ntraj.frame):
            for x1, x2 in zip(ntraj.box[i], atraj.box[i]):
                self.assertAlmostEqual(x1, x2, places=3)
        self.assertEqual(len(nrst.box), 6)
        self.assertEqual(len(arst.box), 6)
        # Make sure the EnergyMinimizerReporter does not fail
        f = StringIO()
        rep = EnergyMinimizerReporter(f, volume=True)
        rep.report(sim)
        rep.finalize()

    def test_private_functions(self):
        """ Tests private helper functions for OMM reporters """
        self.assertEqual(_format_time(7200), (2, 'hr.'))
        self.assertEqual(_format_time(3600), (60, 'min.'))
        self.assertEqual(_format_time(40), (40, 'sec.'))



""" Imports from the Entos ecosystem """

__all__ = ['sierra', 'Molecule', 'QMMMInput', 'HAS_MISTRAL', 'HAS_ENTOS', 'constants']

try:
    import sierra
    from sierra import constants
    from sierra.inputs import Molecule
    try:
        from sierra.inputs import QMMMInput, QMMMSystem
    except ImportError:
        HAS_MISTRAL = False
        QMMMInput = QMMMSystem = None
    else:
        HAS_MISTRAL = True
except ImportError:
    sierra = Molecule = constants = QMMMInput = QMMMSystem = None
    HAS_ENTOS = HAS_MISTRAL = False
else:
    HAS_ENTOS = True



""" Bindings to generate Entos objects from ParmEd objects """
from .converters import to_entos_molecule, to_entos_qmmm_system
from .imports import HAS_ENTOS, HAS_MISTRAL

__all__ = ['to_entos_molecule', 'to_entos_qmmm_system', 'HAS_ENTOS', 'HAS_MISTRAL']



"""
Contains unittests for running OpenMM calculations using native CHARMM files

The energies here should be the same as those defined in the Chamber tests of
test_openmm_amber.py, with a small caveat.  The energies computed in the chamber
tests all used the same (mbondi2) radii.  The radii for the GB models here is
set internally, so there is no way to reproduce the results from the chamber
tests without changing the tests (too time consuming) or changing the CHARMM
parser code in undesirable ways. So some of the nonbonded energies are close to
the analogous values in the Chamber tests, but have been recomputed using ParmEd
with the appropriate radius set.

The explicit solvent energies are also slightly different, since the coordinates
stored in the CHARMM coordinate file are of higher precision than those stored
in the Amber restart file. The energies are very close, but have been
recalculated using ParmEd for comparison here to improve the precision. The
implementation these energies were computed by has already been validated.
"""
import pytest
from parmed.amber.readparm import Rst7
from parmed.charmm import CharmmPsfFile, CharmmCrdFile, CharmmRstFile, CharmmParameterSet
from parmed.exceptions import CharmmWarning, ParameterError
from parmed.openmm.utils import energy_decomposition
from parmed import unit as u, openmm, load_file, UreyBradley
from copy import copy
from math import sqrt
import unittest
from utils import get_fn, TestCaseRelative, mm, app, has_openmm, CPU
import warnings

# System
charmm_gas = CharmmPsfFile(get_fn('ala_ala_ala.psf'))
charmm_gas_crds = load_file(get_fn('ala_ala_ala.pdb'))
charmm_nbfix = CharmmPsfFile(get_fn('ala3_solv.psf'))
charmm_nbfix_crds = CharmmCrdFile(get_fn('ala3_solv.crd'))
charmm_nbfix.box = [3.271195e1, 3.299596e1, 3.300715e1, 90, 90, 90]

# Parameter sets
param22 = CharmmParameterSet(get_fn('top_all22_prot.inp'), get_fn('par_all22_prot.inp'))
param36 = CharmmParameterSet(get_fn('par_all36_prot.prm'), get_fn('toppar_water_ions.str'))

@unittest.skipUnless(has_openmm, "Cannot test without OpenMM")
class TestCharmmFiles(TestCaseRelative):

    def test_gas_energy(self):
        """ Compare OpenMM and CHARMM gas phase energies """
        parm = charmm_gas
        system = parm.createSystem(param22)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=4)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=4)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=4)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=4)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=4)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 9.2210, places=4)

    def test_round_trip(self):
        """ Test ParmEd -> OpenMM round trip with CHARMM gas phase """
        parm = charmm_gas
        system = parm.createSystem(param22)
        self.assertEqual(parm.combining_rule, 'lorentz')
        system2 = openmm.load_topology(parm.topology, system).createSystem()
        con1 = mm.Context(system, mm.VerletIntegrator(0.001), CPU)
        con2 = mm.Context(system2, mm.VerletIntegrator(0.001), CPU)
        con1.setPositions(charmm_gas_crds.positions)
        con2.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, con1)
        energies2 = energy_decomposition(parm, con2)
        self.assertAlmostEqual(energies['bond'], energies2['bond'])
        self.assertAlmostEqual(energies['angle'], energies2['angle'])
        self.assertAlmostEqual(energies['urey_bradley'], energies2['urey_bradley'])
        self.assertAlmostEqual(energies['dihedral'], energies2['dihedral'])
        self.assertAlmostEqual(energies['improper'], energies2['improper'])
        self.assertAlmostEqual(energies['cmap'], energies2['cmap'])
        self.assertRelativeEqual(energies['nonbonded'], energies2['nonbonded'])

    def test_gb1_energy(self): # HCT (uses mbondi radii internally)
        """ Compare OpenMM and CHARMM GB (igb=1) energies """
        parm = charmm_gas
        system = parm.createSystem(param22, implicitSolvent=app.HCT)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -102.1598379, places=5)
        system = parm.createSystem(param22, implicitSolvent=app.HCT,
                                   implicitSolventSaltConc=1.0*u.molar)
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -102.5012873, places=5)

    def test_gb2_energy(self): # OBC1 (uses mbondi2 radii internally)
        """ Compare OpenMM and CHARMM GB (igb=2) energies """
        parm = charmm_gas
        system = parm.createSystem(param22, implicitSolvent=app.OBC1)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -107.8675, places=4)
        system = parm.createSystem(param22, implicitSolvent=app.OBC1,
                                   implicitSolventSaltConc=1.0*u.molar)
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -108.2129, places=4)

    def test_gb5_energy(self): # OBC2 (uses mbondi2 radii internally)
        """ Compare OpenMM and CHARMM GB (igb=5) energies """
        parm = charmm_gas
        system = parm.createSystem(param22, implicitSolvent=app.OBC2)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -103.6186, places=4)
        system = parm.createSystem(param22, implicitSolvent=app.OBC2,
                                   implicitSolventSaltConc=1.0*u.molar)
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -103.9603, places=4)

    def test_gb7_energy(self): # GBn (uses bondi radii internally)
        """ Compare OpenMM and CHARMM GB (igb=7) energies """
        parm = charmm_gas
        system = parm.createSystem(param22, implicitSolvent=app.GBn)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -109.4987850, places=5)
        system = parm.createSystem(param22, implicitSolvent=app.GBn,
                                   implicitSolventSaltConc=1.0*u.molar)
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -109.8465917, places=5)

    def test_gb8_energy(self): # GBn2 (uses mbondi3 radii internally)
        """ Compare OpenMM and CHARMM GB (igb=8) energies """
        parm = charmm_gas
        system = parm.createSystem(param22, implicitSolvent=app.GBn2)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -108.1396, places=4)
        system = parm.createSystem(param22, implicitSolvent=app.GBn2,
                                   implicitSolventSaltConc=1.0*u.molar)
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_gas_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.3351, places=3)
        self.assertAlmostEqual(energies['angle'], 14.1158, places=3)
        self.assertAlmostEqual(energies['urey_bradley'], 0.3669, places=3)
        self.assertAlmostEqual(energies['dihedral'], 14.2773, places=3)
        self.assertAlmostEqual(energies['improper'], 0.3344, places=3)
        self.assertAlmostEqual(energies['cmap'], -0.5239, places=3)
        self.assertRelativeEqual(energies['nonbonded'], -108.4858, places=4)

    def test_dispersion_correction(self):
        """ Compare OpenMM and CHARMM PME energies w/out vdW correction """
        parm = charmm_nbfix
        system = parm.createSystem(param36, nonbondedMethod=app.PME,
                                   nonbondedCutoff=8*u.angstroms,)
        self.assertEqual(parm.combining_rule, 'lorentz')
        for force in system.getForces():
            if isinstance(force, mm.NonbondedForce):
                force.setUseDispersionCorrection(False)
            elif isinstance(force, mm.CustomNonbondedForce):
                force.setUseLongRangeCorrection(False)
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_nbfix_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.1324212, places=4)
        self.assertAlmostEqual(energies['angle'], 1.06880188, places=4)
        self.assertAlmostEqual(energies['urey_bradley'], 0.06142407, places=4)
        self.assertAlmostEqual(energies['dihedral'], 7.81143025, places=4)
        self.assertAlmostEqual(energies['improper'], 0, places=4)
        self.assertAlmostEqual(energies['cmap'], 0.126790, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 6584.01821319, places=4)

    def test_nbfix(self):
        """ Test PME energies of systems with NBFIX modifications """
        parm = charmm_nbfix
        system = parm.createSystem(param36, nonbondedMethod=app.PME,
                                   nonbondedCutoff=8*u.angstroms)
        self.assertEqual(parm.combining_rule, 'lorentz')
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(charmm_nbfix_crds.positions)
        energies = energy_decomposition(parm, sim.context)
        self.assertAlmostEqual(energies['bond'], 1.1324212, places=4)
        self.assertAlmostEqual(energies['angle'], 1.06880188, places=4)
        self.assertAlmostEqual(energies['urey_bradley'], 0.06142407, places=4)
        self.assertAlmostEqual(energies['dihedral'], 7.81143025, places=4)
        self.assertAlmostEqual(energies['improper'], 0, places=4)
        self.assertAlmostEqual(energies['cmap'], 0.126790, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 6514.283116, places=4)

    def test_no_parameters(self):
        """ Test proper error handling when parameters not present """
        parm = CharmmPsfFile(get_fn('ala_ala_ala.psf'))
        parm.rb_torsions.append(parm.dihedrals.pop())
        parm.urey_bradleys.append(UreyBradley(parm.angles[0].atom1, parm.angles[0].atom3))
        self.assertRaises(ParameterError, parm.omm_bond_force)
        self.assertRaises(ParameterError, parm.omm_angle_force)
        self.assertRaises(ParameterError, parm.omm_dihedral_force)
        self.assertRaises(ParameterError, parm.omm_rb_torsion_force)
        self.assertRaises(ParameterError, parm.omm_urey_bradley_force)
        self.assertRaises(ParameterError, parm.omm_improper_force)
        self.assertRaises(ParameterError, parm.omm_cmap_force)

    def test_bad_system(self):
        """ Test error handling of impossible systems """
        parm = copy(charmm_gas)
        parm.load_parameters(param22)
        for d in parm.dihedral_types:
            try:
                for dt in d:
                    dt.scee = 0
            except TypeError:
                d.scee = 0
        self.assertRaises(ValueError, parm.createSystem)

@unittest.skipUnless(has_openmm, "Cannot test without OpenMM")
class TestCharmmDrudeFiles(TestCaseRelative):

    @pytest.mark.xfail
    def test_create_simple_system(self):
        psf = CharmmPsfFile(get_fn("ala3_solv_drude.psf"))
        crd = CharmmCrdFile(get_fn("ala3_solv_drude.crd"))

        params = CharmmParameterSet(get_fn("toppar_drude_master_protein_2013e.str"))

        system = psf.createSystem(params)


"""
This module contains all of the pb namelist variables for the amber
programs and automatically loads those dictionaries with the default
values found in that program (sander or sander.APBS)
"""
from typing import Dict

from .typing import InputDataType

class pb:

    def __init__(self):
        self.sander: Dict[str, InputDataType] = {
            'epsin' : 1.0, 'epsout' : 80.0, 'smoothopt' : 1,
            'istrng' : 0.0, 'pbtemp' : 300.0, 'radiopt' : 1,
            'dprob' : 1.4, 'iprob' : 2.0, 'npbopt' : 0, 'solvopt' : 1,
            'accept' : 0.001, 'maxitn' : 100, 'fillratio' : 2.0,
            'space' : 0.5, 'nbuffer' : 0, 'nfocus' : 2, 'fscale' : 8,
            'npbgrid' : 1, 'arcres' : 0.25, 'bcopt' : 5,
            'scalec' : 0, 'eneopt' : 2, 'frcopt' : 0, 'cutfd' : 5.0,
            'cutnb' : 0.0, 'nsnbr' : 1, 'nsnba' : 1, 'phiout' : 0,
            'phiform' : 0, 'npbverb' : 0, 'npopt' : 2, 'decompopt' : 2,
            'use_rmin' : 1, 'sprob' : 0.557, 'vprob' : 1.3,
            'rhow_effect' : 1.129, 'use_sav' : 1,
            'cavity_surften' : 0.0378, 'cavity_offset' : -0.5692,
            'maxsph' : 400, 'maxarc' : 512, 'cutsa' : 9.0, 'ndofd' : 1,
            'ndosas' : 1, 'fmiccg' : 0.9, 'ivalence' : 1.0,
            'laccept' : 0.1, 'wsor' : 1.9, 'lwsor' : 1.95,
            'pbkappa' : 0, 'radinc' : 0.8, 'expthresh' : 0.2,
            'offx' : 0.0, 'offy' : 0.0, 'offz' : 0.0,
            'sepbuf' : 4.0, 'mpopt' : 0, 'lmax' : 80, 'maxarcdot' : 1500,
            'epsmem': 1.0, 'membraneopt' : 0, 'sasopt' : 0, 'mthick' : 40.0,
            'mctrdz' : 0.0, 'mprob' : 2.7, 'poretype': 1, 'npbverb' : 0,
        }

        self. sanderAPBS: Dict[str, InputDataType] = {
            'dime' : '0,0,0', 'pdime' : '0,0,0', 'cglen' : '0,0,0',
            'fglen' : '0,0,0', 'grid' : '0,0,0', 'nlev' : 4,
            'nonlin' : 0, 'bcfl' : 1, 'nion' : 0, 'pdie' : 2.0,
            'sdie' : 78.4, 'srfm' : 2, 'chgm' : 1, 'temp' : 298.15,
            'gamma' : 0.105, 'sdens' : 10.0, 'calc_type' : 1,
            'calcnpenergy' : 1, 'cmeth' : 1, 'ccmeth' : 1,
            'fcmeth' : 1, 'ionq' : 'NO_DEFAULT',
            'ionc' : 'NO_DEFAULT', 'ionrr' : 'NO_DEFAULT',
            'smvolume' : 10, 'smsize' : 1000.0, 'srad' : 1.4,
            'swin' : 0.3, 'calcenergy' : 2, 'calcforce' : 2,
            'apbs_debug' : 0, 'sp_apbs' : '.false.',
            'apbs_print' : 1, 'wpot' : 0, 'wchg' : 0, 'wsmol' : 0,
            'ispara' : 0, 'radiopt' : 0, 'geom_upd_limit' : 0,
            'evdw_upd_limit' : 0, 'pqr' : 0, 'dime_updates' : 0,
            'wkappa' : 0, 'wdiel' : 0, 'rchg' : 0, 'rkappa' : 0,
            'rdiel' : 0, 'calcnpforce' : 2, 'watompot' : 0,
            'rpot' : 0,
        }



"""
This module contains functionality relevant to loading and parsing DLPOLY CONFIG
(coordinate) files and building a stripped-down Structure from it
"""
from ..constants import TINY
from ..formats.registry import FileFormatType
from ..geometry import box_lengths_and_angles_to_vectors, reduce_box_vectors
from ..utils.io import genopen

class DlpolyConfigFile(metaclass=FileFormatType):
    """ Writes Dlpoly CONFIG files """
    #===================================================

    @staticmethod
    def write(struct, dest, precision=12, combine=False):
        """ Write a Dlpoly CONFIG File from a Structure

        Parameters
        ----------
        struct : :class:`Structure`
            The structure to write to a Dlpoly CONFIG file (must have coordinates)
        dest : str or file-like
            The name of a file or a file object to write the Dlpoly CONFIG file to
        precision : int, optional
            The number of decimal places to print in the coordinates. Default 9
        combine : 'all', None, or list of iterables, optional
            Equivalent to the combine argument of the DlpolyFieldFile.write
            method. If None, system atom order may be changed to meet the need
            for contiguously bonded groups of atoms to be part of a single
            moleculetype. All other values leave the atom order unchanged.
            Default is None.
        """

        def _write_atom_line(atom, atid, resid, has_vels, dest, precision):
            varwidth = 8 + precision
            crdfmt = '%%%d.%df' % (varwidth, precision)
            velfmt = '%%%d.%df' % (varwidth, precision)
            dest.write('%-8s %6d' % (atom.type[:5], atid))
            dest.write('\n')
            dest.write((crdfmt % (atom.xx))[:varwidth])
            dest.write((crdfmt % (atom.xy))[:varwidth])
            dest.write((crdfmt % (atom.xz))[:varwidth])
            dest.write('\n')
            if has_vels:
                dest.write((velfmt % (atom.vx))[:varwidth])
                dest.write((velfmt % (atom.vy))[:varwidth])
                dest.write((velfmt % (atom.vz))[:varwidth])
                dest.write('\n')

        own_handle = False
        if isinstance(dest, str):
            dest = genopen(dest, 'w')
            own_handle = True
        elif not hasattr(dest, 'write'):
            raise TypeError('dest must be a file name or file-like object')

        dest.write('CONFIGURATION\n')
        # CONFIG file key (do we have velocities as well?)
        has_vels = all(hasattr(a, 'vx') for a in struct.atoms)
        if has_vels:
            dest.write('         1')
        else:
            dest.write('         0')
        # Periodic Boundary Key
        if struct.box is not None:
            a, b, c = reduce_box_vectors(*box_lengths_and_angles_to_vectors(*struct.box))
            if all([abs(x-90) < TINY for x in struct.box[3:]]):
                if (a[0] == b[1] and b[1] == c[2]):
                    dest.write('         1')
                else:
                    dest.write('         2')
            else:
                dest.write('         3')
        else:
            dest.write('         0')
        dest.write('     %5d         0.0000000000E+00\n' % len(struct.atoms))
        # Box information
        if struct.box is not None:
            varwidth = 8 + precision
            boxfmt = '%%%d.%df%%%d.%df%%%d.%df\n' % (varwidth, precision, varwidth, precision, varwidth, precision)
            dest.write(boxfmt % (a[0], a[1], a[2]))
            dest.write(boxfmt % (b[0], b[1], b[2]))
            dest.write(boxfmt % (c[0], c[1], c[2]))
        if combine != 'all':
            resid, atid = 0, 0
            # use struct.split to get residue order as per topology file
            split_struct = struct.split()
            n_mols = sum(len(mol[1]) for mol in split_struct)
            unused_atoms = list(struct.atoms)
            for molid in range(n_mols):
                # loop through molids so we can get the correct molecule
                # according to the order they appear
                molecule = [
                    mol[0] for mol in split_struct if molid in mol[1]][0]
                new_molecule = set()  # track atoms added
                last_found_atom = None  # track when CONFIG and FIELD diverge

                for residue in molecule.residues:
                    resid += 1
                    for atom in residue.atoms:
                        # for each atom in split topology get the first
                        # matching occurrence in the original structure
                        for original_atom in unused_atoms:
                            if atom.type == original_atom.type and \
                               atom.name == original_atom.name and \
                               atom.residue.name == original_atom.residue.name:

                                if last_found_atom is not None and \
                                   original_atom.idx != last_found_atom.idx + 1:
                                    # a rearrangement has occurred! Need to do
                                    # extra check that we've found the correct
                                    # original_atom
                                    if len(new_molecule.intersection(
                                            original_atom.bond_partners)) == 0:
                                        # original_atom must be bonded to at
                                        # least one atom in the molecule we
                                        # are currently writing otherwise find
                                        # next candidate
                                        continue

                                atid += 1
                                _write_atom_line(
                                    original_atom, atid % 100000,
                                    resid % 100000, has_vels, dest, precision)
                                new_molecule.add(original_atom)
                                last_found_atom = original_atom
                                unused_atoms.remove(original_atom)
                                break
                        else:
                            raise RuntimeError("Could not find %s" % atom)
        else:
            for atom in struct.atoms:
                resid = (atom.residue.idx + 1) % 100000
                atid = (atom.idx + 1) % 100000
                _write_atom_line(atom, atid, resid, has_vels, dest, precision)

        if own_handle:
            dest.close()
